import Foundation
import Accelerate

class FullyConnectedFilterNode: FilterNode {
    let activation: Activation
    let weight: Weight

    init(source: Node, weight: Weight, activation: Activation = Activation.sigmoid) {
        self.activation = activation

        assert(weight.inputSize == source.outputSize)
        self.weight = weight

        super.init(source: source, outputSize: weight.outputSize)
    }

    override func roll(env: Env) -> Instance {
        let previous = self.source.roll(env: env)
        let instance = Instance(of: self, from: previous)
        vDSP_mmul(weight.slope, 1, previous.output, 1, &instance.output, 1, UInt(outputSize), 1, UInt(inputSize))
        vDSP_vadd(weight.bias, 1, instance.output, 1, &instance.output, 1, UInt(outputSize))
        activation.activate(&instance.output)
        return instance
    }

    override func activationGradient(for output: FloatBuffer) -> FloatBuffer {
        return activation.gradient(output)
    }

    func gradientFilter(source: Node) -> FullyConnectedGradientFilterNode {
        return FullyConnectedGradientFilterNode(source: source, connecting: self)
    }
}

class FullyConnectedGradientFilterNode: FilterNode {
    let connected: FullyConnectedFilterNode
    
    init(source: Node, connecting connected: FullyConnectedFilterNode) {
        self.connected = connected
        
        super.init(source: source, outputSize: connected.inputSize)
    }
    
    override func roll(env: Env) -> Instance {
        let previous = self.source.roll(env: env)
        var connectedInstance: Instance
        if let previous = previous as? FullyConnectedGradientFilterInstance {
            connectedInstance = previous.connected.previous!
        } else {
            connectedInstance = previous.previous!
        }
        var error = connected.activationGradient(for: connectedInstance.output)
        vDSP_vmul(previous.output, 1, error, 1, &error, 1, UInt(previous.output.count))
        var gradient = FloatBuffer(repeating: 0.0, count: connected.weight.slope.count)
        vDSP_mmul(error, 1, connectedInstance.previous!.output, 1, &gradient, 1,
                  UInt(connected.weight.outputSize), UInt(connected.weight.inputSize), 1)
        
        let instance = FullyConnectedGradientFilterInstance(of: self, from: previous,
                                                            connecting: connectedInstance,
                                                            error: error, gradient: gradient)
        vDSP_mmul(error, 1, connected.weight.slope, 1, &instance.output, 1, 1, UInt(outputSize), UInt(error.count))
        return instance
    }
}

class FullyConnectedGradientFilterInstance: Instance {
    let connected: Instance
    
    var error: FloatBuffer
    var gradient: FloatBuffer
    
    init(of node: Node, from previous: Instance?, connecting connected: Instance, error: FloatBuffer, gradient: FloatBuffer) {
        self.connected = connected
        self.error = error
        self.gradient = gradient
        
        super.init(of: node, from: previous)
    }
}
