import Foundation
import Accelerate

typealias LossInput = (expected: FloatBuffer, actual: FloatBuffer)

struct Loss {
    let loss: (_: LossInput) -> Float
    let gradient: (_: LossInput) -> FloatBuffer
    
    static let l2: Loss = Loss(loss: l2Loss, gradient: l2Gradient)
    // crossEntropy + softMax
    // https://deepnotes.io/softmax-crossentropy
    static let crossEntropy: Loss = Loss(loss: crossEntropyLoss, gradient: crossEntropyGradient)
}

func l2Loss(input: LossInput) -> Float {
    var diff = input.expected
    let actual = input.actual
    vDSP_vsub(actual, 1, diff, 1, &diff, 1, UInt(diff.count))
    var result: Float = 0.0
    vDSP_dotpr(diff, 1, diff, 1, &result, UInt(diff.count))
    return result / 2.0
}

func l2Gradient(input: LossInput) -> FloatBuffer {
    var diff = input.expected
    let actual = input.actual
    vDSP_vsub(actual, 1, diff, 1, &diff, 1, UInt(diff.count))
    return diff
}

fileprivate func softmax(input: FloatBuffer) -> FloatBuffer {
    var negmax = -input.max()!
    var buffer = input
    vDSP_vsadd(buffer, 1, &negmax, &buffer, 1, UInt(buffer.count))
    
    buffer = buffer.map{ exp($0) }
    
    var sum: Float = 0.0
    vDSP_sve(buffer, 1, &sum, UInt(buffer.count))
    vDSP_vsdiv(buffer, 1, &sum, &buffer, 1, UInt(buffer.count))
    return buffer
}

func crossEntropyLoss(input: LossInput) -> Float {
    var buffer = softmax(input: input.actual).map{ log($0) }
    let expected = input.expected
    vDSP_vmul(buffer, 1, expected, 1, &buffer, 1, UInt(buffer.count))
    var result: Float = 0.0
    vDSP_sve(buffer, 1, &result, UInt(buffer.count))
    return -result
}

func crossEntropyGradient(input: LossInput) -> FloatBuffer {
    var buffer = softmax(input: input.actual)
    vDSP_vsub(buffer, 1, input.expected, 1, &buffer, 1, UInt(buffer.count))
    return buffer
}
