import Foundation
import Accelerate

protocol Weight {
    var outputSize: Int { get }
    var inputSize: Int { get }

    var slope: FloatBuffer { get }
    var bias: FloatBuffer { get }

    func update(slopeGrads: [FloatBuffer], biasGrads: [FloatBuffer]) -> Void
    func update(slopeGrad: FloatBuffer, biasGrad: FloatBuffer) -> Void
}

extension Weight {
    func update(slopeGrad: FloatBuffer, biasGrad: FloatBuffer) {
        self.update(slopeGrads: [slopeGrad], biasGrads: [biasGrad])
    }
}

fileprivate func mean(vectors: [FloatBuffer]) -> FloatBuffer {
    assert(vectors.count > 0)
    
    var sum: FloatBuffer = vectors.dropFirst().reduce(into: FloatBuffer(vectors.first!)) {
        (sum: inout FloatBuffer, v: FloatBuffer) in
        vDSP_vadd(sum, 1, v, 1, &sum, 1, UInt(sum.count))
    }

    var count = Float(vectors.count)
    vDSP_vsdiv(sum, 1, &count, &sum, 1, UInt(sum.count))

    return sum
}

class InMemoryWeight: Weight {
    let outputSize: Int
    let inputSize: Int
    var learningRate: Float

    var slope: FloatBuffer
    var bias: FloatBuffer

    init(inputSize: Int, outputSize: Int, learningRate: Float,
         slopeRange: Range<Float> = -1.0..<1.0, biasRange: Range<Float> = -1.0..<1.0) {
        self.inputSize = inputSize
        self.outputSize = outputSize
        self.learningRate = learningRate
        
        self.slope = FloatBuffer((0..<inputSize * outputSize).lazy.map{ _ in return Float.random(in: slopeRange) })
        self.bias = FloatBuffer((0..<outputSize).lazy.map{ _ in return Float.random(in: biasRange) })
    }

    func update(slopeGrads: [FloatBuffer], biasGrads: [FloatBuffer]) {
        assert(slopeGrads.count == biasGrads.count)

        let slopeGrad = mean(vectors: slopeGrads)
        let biasGrad = mean(vectors: biasGrads)

        assert(slope.count == slopeGrad.count)
        assert(bias.count == biasGrad.count)

        vDSP_vsma(slopeGrad, 1, &learningRate, slope, 1, &slope, 1, UInt(slope.count))
        vDSP_vsma(biasGrad, 1, &learningRate, bias, 1, &bias, 1, UInt(bias.count))
    }
}
